@@ -1,3 +1,5 @@
+#ifndef LINK_H
+#define LINK_H 1
 #include <time.h>
 
 /* 2 Dimensional Link List */
@@ -25,3 +27,5 @@ extern dir_item *add_right(dir_item *listptr);
 extern dir_item *add_down(dir_item *listptr);
 extern dir_item *free_right(dir_item *listptr);
 extern dir_item *free_dirinfo(dir_item *listptr);
+
+#endif /* !LINK_H */
