@@ -191,7 +191,7 @@ void
 	for (i = 0; i < strlen(progname)-1; i+=1)
 	    printf(" ");
 	
-	fprintf(stdout,
+	printf(
 #ifndef __FreeBSD__
 	       "                  [-c] <[-U <path>] [-u]>\n"
 #else
@@ -243,6 +243,12 @@ void
 	exit(0);
 }
 
+void drop_sgid()
+{
+	if (setgid(GID) == -1)
+	    report_error(FATAL, QUIET, "%s: Could not drop privileges.", progname);
+}
+
 static void
     put_short (int c, FILE *fp)
 {
@@ -281,7 +287,7 @@ int
 	char buf[2];
 	
 	fd = open(database,O_RDONLY);
-
+	
 	if (fd == -1)
 	    return errno;
 	
@@ -342,17 +348,27 @@ void
 	char *part;
 	int i;
 	int res_errno;
-
+	char *tmp_ptr = NULL;
+	int last_sgid = 0;
+	struct stat db_stat;
+	int append;
+	
 	/* Make sure path is not empty */
 	if (!path || strlen(path) == 0) return;
-	
+
+	/* Get the number of already stored paths, if any. */
+	append = 0;
+	while (SLOCATE_PATH && SLOCATE_PATH[append])
+	    append++;
+
 	/* Check how many paths are currently in the string. */
 	i = 1;
 	part = path;
 	while ((part = strchr(part+1, ':'))) i++;
-	
-	/* Allocate enough space to fit existing paths plus new one */
-	SLOCATE_PATH = malloc(i * sizeof(char *));
+
+	/* Allocate enough space to fit existing paths plus new ones plus terminating NULL*/
+	SLOCATE_PATH = realloc(SLOCATE_PATH, (append+i+1) * sizeof(char *));
+
 	if (!SLOCATE_PATH)
 	    report_error(FATAL,QUIET,"%s: parse_decode_path: 'SLOCATE_PATH': malloc: %s\n",progname,strerror(errno));
 	
@@ -365,14 +381,13 @@ void
 	/* Get first path just incase a path with ':'s were specified*/
 	part = strtok(pathcopy, ":");
 
-	i = 0;
 	/* Loop through all paths and validate them.
 	 * use SLOCATE_PATH as an array of pointers to each
 	 * path */
 	while (part) {
 		/* Make sure the path is valid */
 		if (!(res_errno = validate_db(part)))
-		  SLOCATE_PATH[i++] = part;
+		  SLOCATE_PATH[append++] = part;
 		else {
 			if (res_errno == -1)
 			    report_error(WARNING,QUIET,"%s: this is not a valid slocate database: %s\n",progname,part);
@@ -382,9 +397,28 @@ void
 		/* Get next path */
 		part = strtok(NULL, ":");
 	}
-
+	
 	/* Null terminate array */
-	SLOCATE_PATH[i] = NULL;
+	SLOCATE_PATH[append] = NULL;
+	
+	/* Sort sgid slocate db's to the top */
+	for (i = 0; SLOCATE_PATH[i]; i++) {
+		if (stat(SLOCATE_PATH[i], &db_stat) == -1)
+		    report_error(FATAL, QUIET, "%s: Could not stat DB: %s: %s\n", progname, SLOCATE_PATH[i], strerror(errno));
+		
+		if (db_stat.st_gid != SLOC_GID)
+		    continue;
+		
+		if (i != last_sgid) {
+			tmp_ptr = SLOCATE_PATH[last_sgid];
+			SLOCATE_PATH[last_sgid] = SLOCATE_PATH[i];
+			SLOCATE_PATH[i] = tmp_ptr;
+		}
+		
+		last_sgid += 1;
+		
+	}
+
 }
 
 /* Parse Dash */
@@ -996,7 +1030,7 @@ int
 	if (!NEWOUTPUT && UID != SLOC_UID) {
 		report_error(FATAL,QUIET,"%s: You are not authorized to create a default slocate database!\n",progname);
 	}
-
+		
 	fd = fopen(TMPSLOCATEDB,"w");
 
 	if (!fd)
@@ -1144,7 +1178,7 @@ int
 	int ptr_offset;
 	char one_char[1];
 	char *begin_ptr;
-        int begin_offset=0;
+	int begin_offset=0;
 	int tot_size = MIN_BLK;
 	int cur_size;
 	int code_tot_size = MIN_BLK;
@@ -1154,6 +1188,21 @@ int
 	char *cp=NULL;
 #endif
 	char *bucket_of_holding=NULL;
+        gid_t cur_gid;
+	struct stat db_stat;
+	
+	cur_gid = getegid();
+	
+	if (stat(database, &db_stat) == -1)
+	    report_error(FATAL, QUIET, "%s: Could not stat '%d': %s\n", progname, strerror(errno));
+
+	/* If the database's file group is not apart of the 'slocate' group,
+	 * drop privileges. When multiple databases are specified, the ones
+	 * apart of the 'slocate' group will be searched first before the
+	 * privileges are dropped. */
+
+	if (cur_gid == SLOC_GID && db_stat.st_gid != SLOC_GID)
+	    drop_sgid();
 
 	if ((fd = open(database,O_RDONLY)) == -1) {
 		report_error(WARNING,QUIET,"%s: decode_db(): %s: %s\n",progname,database,strerror(errno));
@@ -1238,7 +1287,7 @@ int
 
 		cur_size += bytes;
 
-		code_num = (short)*begin_ptr;
+		code_num = (short) *((signed char *) begin_ptr);
 		begin_ptr += 1;
 
 		if (code_num == SLOC_ESC) {
@@ -1246,19 +1295,10 @@ int
 		} else if (code_num > 127)
 		    code_num = code_num - 256;
 
-		/* FIXME sometimes pathlen is < 0 but it shouldn't be.
-		 * corrupt database file? 
-		 * This could be from a bug in frcode() or decode_db(). I
-		 * am leaning toward frcode() at the moment */
-
 		code_ptr += code_num;
+
 		pathlen = code_ptr - codedpath;
-		
-		if (pathlen < 0) {
-			fprintf(stderr,"%s: decode_db() aborted. Corrupt database?\n",progname);
-			exit(1);
-		}
-		
+
 		jump = 0;
 		while (!jump) {
 			ch = *begin_ptr;
@@ -1407,14 +1447,17 @@ int
 	/* Print usage if not enough arguments were specified */
 	if (args < 2 && !ROOTDIR)
 	    usage();
-	
+
 	/* Get user IDs */
-	UID = getuid();        
+	UID = getuid();
 	GID = getgid();
 
+	/* Get the 'slocate' group GID */
+	SLOC_GID = get_gid(GRPFILE);
+
 	/* Add the LOCATE_PATH environment variable to the list of databases to search in */
 	parse_decode_path(getenv("LOCATE_PATH"));
-	
+
 	/* Loop through specified command line arguments */
 	while ((ch = getopt(args,argv,"VvuhqU:r:o:e:l:d:-:n:f:ci")) != EOF) {
 		switch(ch) {
@@ -1458,6 +1501,10 @@ int
 			break;
 		 /* Update the default database */
 		 case 'u':
+			/* Always drop privileges when creating a database. This way, users created DB's won't be
+			 * owned by the slocate group since the slocate binary is SGID */
+			drop_sgid();
+			
 			/* Check to make sure default database directory exists */
 			check_dir(SLOCATEDB_DIR);
 			ROOTDIR=1;
@@ -1466,6 +1513,7 @@ int
 			break;
 		 /* Update a specified database */
 		 case 'U':
+			drop_sgid();
 			ADD_SLOCATEDB = 0;
 			ROOTDIR=0;
 			SPECDIR=1;
@@ -1537,10 +1585,6 @@ int
 	if (ADD_SLOCATEDB)
 	    parse_decode_path(SLOCATEDB);
 
-	/* Get the 'slocate' group GID */
-	SLOC_GID = get_gid(GRPFILE);
-	
-	
 	/* if the -U option has been used, start to create the database at specified path */
 	if (SPECDIR)
 	    res = create_db(spec_dir);
